/* Module support (additional routines) for GNU Pascal
   Copyright (C) 1997-2001 Free Software Foundation, Inc.

This file is part of GNU Pascal.

GNU Pascal is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GNU Pascal is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Pascal; see the file COPYING. If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA. */

/*
 * Authors: Alexei Volokhov <voh@ispras.ru>
 *          Peter Gerwinski <peter@gerwinski.de>
 */

#include "gbe.h"

#include "gpc-defs.h"
#include "util.h"
#include "module.h"
#include "gpi-hash.h"

/* In this module, we are using obstacks separately from the GBE.
 */
#ifdef EGCS97
# include "obstack.h"
#endif
#define obstack_chunk_alloc xmalloc
#define obstack_chunk_free free
extern void gcc_obstack_init PARAMS ((struct obstack *obstack));

#ifdef DEBUG_HASH
#define DEBUG_HASH_KEY (int) 0x54637281
#endif

/* Symbolic constants for the "lseek" function: */
#ifndef SEEK_SET
# define SEEK_SET 0      /* Set file pointer to "offset" */
#endif
#ifndef SEEK_CUR
# define SEEK_CUR 1      /* Set file pointer to current plus "offset" */
#endif
#ifndef SEEK_END
# define SEEK_END 2      /* Set file pointer to EOF plus "offset" */
#endif

/* Circle buffer routines */
#define T void *

struct circle_buf
{
  /* contents */
  T * items;
  /* available size */
  unsigned long a_size;
  /* first input object */
  unsigned long fio;
  /* last input oject */
  unsigned long lio;
  /* allocation stack */
  struct obstack * buf_obstack;
};

#define AROUND_SIZE(SIZE,BY) (((SIZE) + ((1 << (BY)) -1)) << (BY) >> (BY))

/* Initialize procedure.
   buf_obstack must be already initialized.
   start_size must be >0 */

static void
init_circle_buf (self,
                 buf_obstack,
                 start_size)
     struct circle_buf * self;
     struct obstack * buf_obstack;
     unsigned long start_size;
{
  self->a_size = AROUND_SIZE (start_size, 2);
  self->buf_obstack = buf_obstack;
  self->items = obstack_alloc (self->buf_obstack, self->a_size * sizeof (T));
  memset ((void *) self->items, 0, self->a_size * sizeof (T));
  self->fio = self->lio = 0;
}

static void
realloc_circle_buf (self)
     struct circle_buf * self;
{
  /* size is grown to old_size *2 */
  unsigned long old_size = self->a_size;
  T * new_items;
  self->a_size = AROUND_SIZE (self->a_size << 1, sizeof (T));
  new_items =  obstack_alloc (self->buf_obstack, self->a_size * sizeof (T));
  /* buffer full lio == fio here */
  if (old_size > self->fio)
    memcpy ((void *) &new_items[0], (void *) &(self->items[self->fio]),
            (old_size - self->fio) * sizeof (T));
  if (self->lio > 0)
    memcpy ((void *) &new_items[old_size - self->fio], (void *) &(self->items[0]),
            self->lio * sizeof (T));
  self->lio = old_size;
  self->fio = 0;
  /* free buf */
  self->items = new_items;
}

/* Insert obj into queue */
static void
in_circle_buf (self, obj)
     struct circle_buf * self;
     T obj;
{
  /* empty buffer */
  if (self->fio == self->lio)
    {
      self->fio = 0;
      self->lio = 0;
    }

  self->items [self->lio] = obj;
  self->lio ++;

  /* not empty circle buffer */
  if (self->lio >= self->a_size)
    self->lio = 0;
  /* full buffer ==> realloc it */
  if (self->lio == self->fio)
    {
      realloc_circle_buf (self);
    }
}

/* Get object from queue.
  if no objects in the buffer returns 0 and set no_more = 1 */
static T
out_circle_buf (self, no_more)
     struct circle_buf * self;
     int * no_more;
{
  T obj;
  if (self->fio == self->lio)
    {
      *no_more = 1;
      return 0;
    }
  obj = self->items[self->fio++];
  if (self->fio == self->lio)
    {
      self->fio = 0;
      self->lio = 0;
    }
  else if (self->fio >= self->a_size)
    self->fio = 0;
  return obj;
}

/* Returns 1 if buffer is empty */
static int
is_empty_buf (struct circle_buf * self)
{
  return self->fio == self->lio;
}

/* Crear buffer */
static void
free_circle_buf (struct circle_buf * self)
{
  self->fio = self->lio = 0;
}

/* integer type for void * representative */
#define IRP_TYPE long

struct uid_ptr
{
  tree node;
  int uid;
  struct uid_ptr * next;
};

#define UID_IS_NOT_SAVED(UID) ((UID) > 0)
#define SAVE_UID(UID) (-(UID))
#define NATIVE_UID(UID) ((UID) > 0 ? (UID) : -(UID))

#define MAX_HASH_TABLE 8191
static struct uid_ptr *hash_table[MAX_HASH_TABLE];  /* id hash buckets */

off_t * offs_array;
long offs_array_size;
struct obstack imbed_file_obstack;
int imbed_file_obstack_inited = 0;
char * imbed_firstobj;
long uid_counter;

struct circle_buf * cbf;

/* alloc and clear memory */
static void *
imb_alloc (size)
     size_t size;
{
  void * tmp;
  tmp = obstack_alloc (&imbed_file_obstack,size);
  if (size)
    memset (tmp, 0, size);
  return tmp;
}

/* init data for store gpi file */
static void
init_outbed (a_size)
     int a_size;
{
  uid_counter = 0;
  if (!imbed_file_obstack_inited)
    {
      gcc_obstack_init (&imbed_file_obstack);
      imbed_file_obstack_inited = 1;
      imbed_firstobj = (char *) imb_alloc (0);
      cbf = (struct circle_buf *) imb_alloc (sizeof (struct circle_buf));
      init_circle_buf (cbf, &imbed_file_obstack, a_size);
    }
  else
    {
      if (cbf)
        free_circle_buf (cbf);
      else
        {
          cbf = (struct circle_buf *)
            imb_alloc (sizeof (struct circle_buf));
          init_circle_buf (cbf, &imbed_file_obstack, a_size);
        }
    }
  offs_array_size = a_size;
  offs_array = imb_alloc (offs_array_size * sizeof (off_t));
}

/* We must avoid negative values.  Naturally, we would cast this to
 * `unsigned ptrdiff_t', but since we cannot be sure that the platform
 * we compile GPC on will support this, we use `abs' instead.
 */
#define HASH_FUNC(NODE) ((abs ((NODE) - (tree) 0) >> 2) % MAX_HASH_TABLE)

/* return pointer to a node uid field */
static int *
get_uid (node)
     tree node;
{
  IRP_TYPE hi;
  struct uid_ptr *l;

  hi = HASH_FUNC (node);
  /* Search table for node */
  for (l = hash_table[hi]; l; l =l->next)
    if (l->node == node)
      return &l->uid;
  return 0;
}

/* set unique identifier to a node
   uid == 0 means next available.
   return pointer to uid field */
static int *
set_uid (node, uid)
     tree node;
     int uid;
{
  IRP_TYPE hi;
  struct uid_ptr *l;

  if (uid)
    {
      assert (uid > uid_counter);
      uid_counter = uid;
    }
  else
    uid = ++uid_counter;
  if (uid_counter >= offs_array_size)
    {
      long new_size =  offs_array_size << 2;
      off_t * new_offs_array = imb_alloc (new_size * sizeof (off_t));
      memcpy (new_offs_array, offs_array, offs_array_size*sizeof (off_t));
      offs_array = new_offs_array;
      offs_array_size = new_size;
    }
  /* Compute hash code */
  hi = HASH_FUNC (node);
  l = (struct uid_ptr *) imb_alloc (sizeof (struct uid_ptr));
  l->node = node;
  l->uid = uid;
  l->next = hash_table[hi];
  hash_table[hi] = l;
  return &l->uid;
}

/* Check/or set uid for node and  add node to offs_array or circular buf*/
int
hashed_save (node, s, depth)
     tree node;
     FILE *s;
     int depth;
{
  int * uid_p;
  int native_uid;

  if (!node)
    {
      native_uid = 0;
      fwrite1 (&native_uid, sizeof (native_uid), s);
      return 1;
    }
  uid_p = get_uid (node);
  if (!uid_p)
    uid_p = set_uid (node, 0);
  if (UID_IS_NOT_SAVED(*uid_p)) /* not saved yet */
    {
      if (!depth)
        {
          offs_array [*uid_p] = ftell (s);
#ifdef DEBUG_HASH
          {
            int key = DEBUG_HASH_KEY;
            fwrite1 (&key, sizeof (key), s);
          }
#endif
          *uid_p = SAVE_UID(*uid_p);
          return 0;
        }
      else
        {
          fwrite1 (uid_p, sizeof (*uid_p), s);
          in_circle_buf (cbf, node);
        }
    }
  else
    {
      native_uid = NATIVE_UID(*uid_p);
      fwrite1 (&native_uid, sizeof (native_uid), s);
    }
  return 1;
}

/* store all nodes from circular buf */
void
flush_suspended (f)
     FILE * f;
{
  int * uid_p;
  tree node;
  int no_more;
  while (!is_empty_buf(cbf))
    {
      node = out_circle_buf (cbf, &no_more);
      uid_p = get_uid (node);
      assert (uid_p);
      assert (*uid_p);
      if (uid_p && *uid_p > 0)
        store_tree (node, f, 0);
    }
}

off_t header_off;
struct imbed_header
{
  off_t offs_array_off;
  long offs_array_size;
#ifdef DEBUG_HASH
  int key;
#endif
} ss;

/* Save place for header
   set UID for builtin nodes */
void
store_header (s, nn)
     FILE * s;
     int nn;
{
  int * uid_p;
  header_off = ftell (s);
#ifdef DEBUG_HASH
  ss.key = 0;
#endif
  fwrite1 (&ss, sizeof (ss), s);
  memset (hash_table, 0, MAX_HASH_TABLE * sizeof (struct uid_ptr *));
  init_outbed (nn + 1024);
#define PROCESS_NODE(n,node)  \
   uid_p = set_uid (node, n);  \
   *uid_p = -*uid_p;
#include "nodes.def"
#undef PROCESS_NODE
}

/* Write UID OFFSET table */

void
update_header (s)
     FILE * s;
{
  struct imbed_header ss;
  ss.offs_array_off = ftell (s);
  ss.offs_array_size = offs_array_size;
#ifdef DEBUG_HASH
  ss.key = -DEBUG_HASH_KEY;
#endif
  fseek (s, header_off, SEEK_SET);
  fwrite1 (&ss, sizeof (ss), s);
  fseek (s, 0, SEEK_END);
  fwrite1 (offs_array, offs_array_size * sizeof (off_t), s);
}

off_t * imbed_offs_array;
tree * loaded_nodes;

/* Read UID OFFSET table */
void
load_module_header (s)
     MEMFILE * s;
{
  off_t cp;
  memset (hash_table, 0, MAX_HASH_TABLE * sizeof (struct uid_ptr *));
  if (!imbed_file_obstack_inited)
    {
      gcc_obstack_init (&imbed_file_obstack);
      imbed_file_obstack_inited = 1;
      imbed_firstobj = (char *) imb_alloc (0);
    }
  mread1 (&ss, sizeof (ss), s);
#ifdef DEBUG_HASH
  assert (ss.key == -DEBUG_HASH_KEY);
#endif
  imbed_offs_array = imb_alloc (ss.offs_array_size * sizeof (off_t));
  loaded_nodes = (tree *) imb_alloc (ss.offs_array_size * sizeof (tree));
  cp = mtell (s);
  mseek (s, ss.offs_array_off);
  mread1 (imbed_offs_array, ss.offs_array_size * sizeof (off_t), s);
  mseek (s, cp);
#define PROCESS_NODE(n,node) loaded_nodes[n] = node
#include "nodes.def"
#undef PROCESS_NODE
}

/* Check whether the node with UID is already loaded.
   If it is returns node
   If not seek file for reading the node and returns 0 */
tree
hashed_load (s, uid)
     MEMFILE * s;
     int uid;
{
  if (loaded_nodes[uid])
    return loaded_nodes[uid];
  mseek (s, imbed_offs_array[uid]);
#ifdef DEBUG_HASH
  {
    int key;
    mread1 (&key, sizeof (key), s);
    assert (key == DEBUG_HASH_KEY);
  }
#endif
  return 0;
}

/* Mark UID as loaded node */
tree
hash_loaded_node (node, uid)
     tree node;
     int uid;
{
  assert (node);
  assert (uid);
  /* Loops are possible */
  if (loaded_nodes[uid])
    {
      assert (TREE_CODE (loaded_nodes[uid]) == TREE_CODE (node));
      return loaded_nodes[uid];
    }
  return ((loaded_nodes[uid] = node));
}

void
clean_imbed_file_obstack ()
{
  imbed_file_obstack_inited = 0;
  obstack_free (&imbed_file_obstack, imbed_firstobj);
}
